from setuptools import setup

from setuptools import setup
setup(name='causaln',
version='1.0',
description='for defining causal networks and generating example data',
url='https://github.com/PeterJamesWinn',
author='Peter James Winn',
license='MIT',
packages=['causaln'],
zip_safe=False)
